#!/usr/bin/python
#
# Cerebro Project
#
#
# Polychronis Ypodimatopoulos <ypod@mit.edu>
#
# February 2008
# Viral Comm - MIT Media Lab
#

import sys, time, os, threading
from string import upper
from random import random
from struct import pack, unpack
from math   import ceil, pi
from select import select
from socket import *
from cerebro.common import *

QUESTION = "What's the frequency Kenneth?"

class MeshPresenceInterface(threading.Thread):
	def __init__(self, interval=1, iface="", port = 44555, otherport = 33444):
		threading.Thread.__init__(self)
		#sys.stdout = open('/var/log/cerebro', 'w', 0)
		#sys.stderr = sys.stdout
		
		self.interval	= interval
		self.go			= 1
		self.ETH_PRESENCE	= 0x7602

		self.SYSTEM = self.guess_system()
		
		(self.iface, self.hostname) = self.setup_iface(iface)
		
		self.rawSock	= socket(AF_PACKET, SOCK_RAW, htons(self.ETH_PRESENCE))
		self.rawSock.bind((self.iface, self.ETH_PRESENCE))
		self.binary_mac	= self.rawSock.getsockname()[4]
		self.src_mac	= self.rawSock.getsockname()[4]
		self.dst_mac	= chr(255)*6

		self.UDPsock = socket(AF_INET,SOCK_DGRAM)
		host = "localhost"
		
		self.other_addr = (host,otherport)
		self.local_addr = (host,port)
		self.UDPsock.bind(self.local_addr)
		
		# sources keeps track of the last timestamp from each source, so that
		# replicate frames are dropped.
		self.sources = {}
		
	def guess_system(self):
		plat = sys.platform
		if plat == "linux2":
			if os.popen("grep OLPC /etc/issue").readline().strip():
				return "OLPC"
			
			if os.popen("grep -i Gumstix /etc/issue").readline().strip():
				return "GUMSTIX"
			
			if os.popen("grep -i ubuntu /etc/issue").readline().strip():
				return "GENERIC"
			
			if os.popen("grep -i N800 /etc/issue").readline().strip():
				return "N800"
			
			try:
				lintype = file("/proc/net/wireless", "r").readlines()
			except:
				print "PANIC: Failed to read /proc/net/wireless"
				
			# /proc/net/wireless must have more than 2 lines to find a wireless interface
			if len(lintype) < 3:
				print "PANIC: No wireless interface found in /proc/net/wireless"
				
		elif plat == "win32":
			pass
		
		elif plat == "darwin":
			return "APPLE"
		
		else:
			print "PANIC: Unknown platform", plat
			sys.exit()
		
	def setup_iface(self, iface):
		res      = ""
		hostname = ""
		
		if self.SYSTEM == "GUMSTIX":
			res = "mwlan0"
			promisc = os.popen("ifconfig %s promisc" % res).readline().strip()
			hostname = os.popen("hostname").readline().strip()
			
		elif self.SYSTEM == "OLPC":
			res = "msh0"
			if iface: res = iface
			ttl = os.popen("iwpriv msh0 mesh_set_ttl 1").readline().strip()
			hostname = os.popen("grep nickname /home/olpc/.sugar/default/config|awk '{print $3}'").readline().strip()
			#promisc = os.popen("/sbin/iwconfig %s mode ad-hoc channel 1 essid '%s:I hate essids!'" % (res, hostname)).readline().strip()
			promisc = os.popen("/sbin/iwconfig %s mode ad-hoc channel 1 essid 'cerebro'" % res).readline().strip()
			#promisc = os.popen("/sbin/iwconfig %s channel 1" % res).readline().strip()
			promisc = os.popen("/sbin/ifconfig %s -promisc up" % res).readline().strip()
		elif self.SYSTEM == "APPLE":
			res = "en0"
		else:
			res = iface
			promisc = os.popen("/sbin/iwconfig %s mode ad-hoc channel 1 essid 'cerebro'" % res).readline().strip()
			#promisc = os.popen("/sbin/ifconfig %s promisc up" % res).readline().strip()
			hostname = os.popen("hostname").readline().strip()
			
		if iface != "lo":
			#channel = os.popen("/sbin/iwconfig %s mode ad-hoc channel 1" % res).readline().strip()
			time.sleep(2)
			
		return (res, hostname)
		
		
		
	def makeFrame(self, dst_mac, src_mac, ftype, load, burst=1):
		nodeid    = self.hostname
	
		if len(dst_mac) == 17:
			dst_mac = "".join([chr(int(byte, 16)) for byte in dst_mac.split(':')])
		
		if len(src_mac) == 17:
			src_mac = "".join([chr(int(byte, 16)) for byte in src_mac.split(':')])
		
		if len(dst_mac) != 6 or len(src_mac) != 6:
			print "PANIC: Bad address: You have a bug somewhere!!!"
			print "DST_MAC: %d       SRC_MAC: %d" % (len(dst_mac), len(src_mac))
			sys.exit()
			
		if len(nodeid) < 10: nodeid += "          "
		
		output = ""
		output += dst_mac 					# destination	(6 bytes)
		output += src_mac 					# source		(6 bytes)
		output += pack("!H", self.ETH_PRESENCE)	# type 			(2 bytes)
		output += nodeid[:10]				# nodeId		(10 bytes)
		output += pack("I", ceil(time.time()))	# timestamp
		output += pack("H", len(load))		# load length	(2 bytes)
		output += load
		return output

	def parse_frame(self, data):
		ftype	  = data[14]
		#nodeid	  = data[15:25]
		dst       = data[0:6]
		src       = data[6:12]
		nodeid    = data[14:24]
		tstamp    = unpack("I", data[24:28])[0]
		loadlen   = unpack("H", data[28:30])[0]
		load      = data[30:30+loadlen]
			
		return src, nodeid, load, tstamp
		
		
	def run(self):

		print "System type:    ", self.SYSTEM
		print "Using interface:", self.iface
		print "Hostname:       ", self.hostname
		
		while self.go:
			### WAIT FOR A FRAME TO ARRIVE FROM SOME NEIGHBOR
			ins, outs, exc = select([self.UDPsock, self.rawSock], [], [], self.interval)

			### IF SOMETHING ARRIVES
			for s in ins:
				
				if s == self.UDPsock:
					#print "it's from local"
					data, addr = self.UDPsock.recvfrom( 2000 )
					if data == QUESTION:
						print "question was asked"
						if len(self.binary_mac) != 6:
							self.binary_mac = "fakmac"
						print "MPI: my mac is", bin2mac(self.binary_mac)
						self.UDPsock.sendto(self.binary_mac, self.other_addr)
						continue
					#print "data", len(data)
					frame = self.makeFrame(self.dst_mac, self.src_mac, self.ETH_PRESENCE, data)
					#print "frame", len(frame)
					sent = self.rawSock.send( frame )
					#print "sent", sent
				
				if s == self.rawSock:
					#print "it's from remote"
					try:
						(data, addr)		= self.rawSock.recvfrom( 2000 )
					except:
						print "PANIC: Trying to receive: Network down?"
						raise
					(src, nodeid, load, tstamp)	= self.parse_frame( data )
					
					if src in self.sources:
						if tstamp > self.sources[src]:
							self.sources[src] = tstamp
						else:
							print "Dropped duplicate frame because of MAC ACK that was not received"
							print tstamp
							print self.sources
							break
					else:
						self.sources[src] = tstamp
					
					if (self.UDPsock.sendto(src+load, self.other_addr)):
						pass
						#print "instance sent:", data
		print "MeshPresenceInterface exiting..."

if __name__== "__main__":
	if len(sys.argv) < 2:
		interface = "ath0"
	else:
		interface = sys.argv[1]
		
	try:
		ps = MeshPresenceInterface(iface=interface)
		ps.run()
	except KeyboardInterrupt:
		print "exiting"
