# Copyright (C) 2008, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from time import time

class MST:
    def __init__(self):
        self.node = []
        self.dist = []
        self.path = []
        self.seqn = []
    def set_path(self, n, d, p, s):
        #if len(p) < 2:
            #print "MST: error inserting path:", p,"Too small!"
            #return
        try:
            i = self.node.index(n)
            self.dist[i] = d
            self.path[i] = p
            self.seqn[i] = s
        except:
            self.node.append(n)
            self.dist.append(d)
            self.path.append(p)
            self.seqn.append(s)
    def del_path(self, n):
        try:
            i = self.node.index(n)
            del self.node[i]
            del self.dist[i]
            del self.path[i]
            del self.seqn[i]
        except:
            print "MST: error trying to delete path to", n
    def get_path(self, n):
        try:
            i = self.node.index(n)
            return self.path[i]
        except:
            print "MST: error trying to retrieve path to", n
    def get_dist(self, n):
        try:
            i = self.node.index(n)
            return self.dist[i]
        except:
            print "MST: error trying to retrieve distance to", n
    def __iter__(self):
        tmp = []
        for i in xrange(len(self.path)):
            yield self.node[i], self.path[i], self.dist[i], self.seqn[i]
    def clear(self):
        self.node = []
        self.dist = []
        self.path = []
        self.seqn = []
        
class Node:
    def __init__(self, nid, par=None):
        self.nid         = nid
        self.goto        = {}
        self.parent      = par
        self.dist        = 0
        self.seqn        = 0
        self.arr_est     = 0
        self.arr_times   = [time()]
        self.total_dist  = 0
        self.last_update = time()
    def add_node(self, nodeid, cost, seqn):
        node = Node(nodeid, par=self)
        node.dist = cost
        node.seqn = seqn
        self.goto[nodeid] = node
    def __iter__(self):
        for k,path in self.preorder():
            yield k
    def __str__(self):
        #for c in self.nid:
            #if ord(c) < 32 or ord(c) > 126:
                #return ':'.join(["%02X" % ord(c) for c in self.nid])
        return self.nid
    def postorder(self):
        for x in self.goto.values():
            for k in x.postorder():
                yield k
            yield x
    def preorder(self, path=[], cost=0):
        # The following was a horrible choice! "Yield" is not a recursive call
        # and "path" will keep accumulating: "path" does not expire 
        # as you exit this function because of "yield"
        #path += [self.nid]
            
        for x in self.goto.values():
            x.total_dist = x.dist + self.total_dist
            yield x, path + [x.nid]
            for (k, _path) in x.preorder(path + [x.nid]):
                yield k, _path
    def ancestors(self):
        cur = self
        while cur.parent:
            yield cur.parent
            cur = cur.parent
    def has_parent(self):
        return self.parent
    def add_arrival(self):
        self.arr_times.append(time())
    def follow_path(self, path):
        """ Calculates the weight of a given path from current node """
        cur = self
        w = 0
        for n in path:
            try:
                if cur.nid == n:
                    continue
                cur = cur.goto[n]
                w += cur.DIST  ################################################### ADD WEIGHTS?????
            except:
                print "PANIC: weight: Cannot goto %s from %s path:" % (n, cur), path
                raise
        return cur, w
    #def build_mindists(self):
        #for n, path in self.preorder():
            #parent_weight = 0
            #if n.has_parent():
                #parent_weight = n.parent.total_dist
            #n.total_dist = n.dist + parent_weight  ############################### ADD WEIGHTS?????
            
        #for n, path in self.preorder():
            #if not n.nid in m.node or n.total_dist < m.get_dist(n.nid):
                #m.set_path(n=n.nid, d=n.total_dist, p=path)
    def get_weight(self, path):
        """ Calculates the weight of a given path from current node """
        cur = self
        w = 0
        for n in path:
            try:
                cur = cur.goto[n]
                w += cur.DIST  ################################################### ADD WEIGHTS?????
            except:
                self.pout(0, "PANIC: weight: Cannot goto %s from %s" % (n, cur))
        return cur, w
    
    
if __name__== "__main__":
    m=MST()
    a = Node("pol")

    def followpath(path):
        cur = a
        for n in path:
            try:
                cur = cur.goto[n]
            except:
                print "followpath: Could not go to", n
        return cur

    d = 1
    for k in "vwxyz":
        a.add_node( k, d )
        d += 1
    
    for n in a.goto.values():
        n.add_node( "%s%s1"%(n.nid,n.nid), d )
        n.add_node( "%s%s2"%(n.nid,n.nid), d )
        #n.add_node( Node(n.nid+n.nid+"3"), d )
        #n.add_node( Node(n.nid+n.nid+"4"), d )
        d += 1
        
    for n in a.goto.values():
        for k in n.goto.values():
            k.add_node( "%s%s%s1"%(n.nid,n.nid,n.nid), d )
            k.add_node( "%s%s%s2"%(n.nid,n.nid,n.nid), d )
            d += 1
        
        
        
    for n,p in a.preorder():
        print n, p, n.dist, n.total_dist
        #if n.goto
        
    for n in a.postorder():
        parent_weight = n.parent.total_dist
        n.total_dist = n.dist + parent_weight  ############################### ADD WEIGHTS?????
    
    for k in a.postorder():
        print "before:", k
        del k
    
    a.goto.clear()
    print a
    print a.goto
    #for n in a.postorder():
        #print n, n.dist, n.total_dist
    
    #tt = followpath(['v', 'vv1', 'vv1vv1vv11'])
    #print tt
    
    #path = [p.nid for p in tt.ancestors()]
    #print path
    
    #path = []
    #for p in tt.ancestors():
        #path.insert(0, p.nid)

    #print path
    
    
    #a.build_mindists()
    #for p in m:
        #print p
    
