#!/usr/bin/python

"""
graph.py

Layouts a graph based on fixed distance and electro repulsion.

The layout algorithm is dynamic - needs more than one iteration.
Use iterate_layout(nbrIterations,NodeArray,self.DistanceMatrix,Width,Height) and specify the number of iteration - nbrIterations.

You can modify the consts K (repulsive force) and C (force weight) to achieve desired results in less iterations.

The functions are based on a Node data structure that needs to have *at least* a similar structure as the following

struct Node
{
	double _x,_y;		//position
	double _fx,_fy;	//force
}

David Gauthier - gauthier@media.mit.edu

Example: iterate_layout(10,NodeArray,self.DistanceMatrix,800,800);
"""

from math import pi, cos, sin, sqrt

class Pos:
	_x  = 0
	_y  = 0
	_fx = 0
	_fy = 0

class Graph:
	def __init__(self, nbrIterations=10):
		self.K = 20
		self.C = 0.1
		self.DEBUG = 0
		#self.DistanceMatrix = DistanceMatrix
		self.nbrIterations = nbrIterations
		self.Width = 0
		self.Height = 0
		self.DistanceMatrix = []
		self.NodeArray = []
		for i in xrange(len(self.DistanceMatrix)):
			self.NodeArray.append(Pos())
		
	def get_layout(self, DistanceMatrix, edge):
		self.Width          = edge
		self.Height         = edge
		self.DistanceMatrix = DistanceMatrix
		self.NodeArray      = []
		for i in xrange(len(self.DistanceMatrix)):
			self.NodeArray.append(Pos())
		
		self.NodeArray[0]._x = self.Width/2
		self.NodeArray[0]._y = self.Height/2
		self.place(0)
		
		for j in xrange(self.nbrIterations):
			self.layout()
			
		return self.NodeArray
	
	def place(self, rowIndex):
		if rowIndex >= len(self.NodeArray):
			return
		
		cnt = 0
		
		i = rowIndex+1
		while i < len(self.NodeArray):
			if self.DistanceMatrix[rowIndex][i] != 0:
				cnt = cnt + 1
			i += 1
		
		if cnt == 0:
			return
		
		interval = 2*pi/cnt
		
		cc = 0
		
		j = rowIndex+1
		while j < len(self.NodeArray):
			if self.DistanceMatrix[rowIndex][j] != 0:
				cc = cc + 1
				self.NodeArray[j]._x = self.DistanceMatrix[0][j] * cos(cc*interval) + self.NodeArray[rowIndex]._x  #polar to cartesian
				self.NodeArray[j]._y = self.DistanceMatrix[0][j] * sin(cc*interval) + self.NodeArray[rowIndex]._y  #polar to cartesian
				self.place(j)
			j += 1
	
	def layout(self):
		self.relax()
		
		for i in xrange(len(self.NodeArray)):
			if (self.DEBUG): print "------------------------------------------"
			if (self.DEBUG): print "Node -", i
		
			vertextmp = self.NodeArray[i]
			row = self.DistanceMatrix[i]
		
			if (self.DEBUG): print "c =", self.C, "fx =", self.NodeArray[i]._fx, "fy =", self.NodeArray[i]._fy
				
			xmove = self.C * vertextmp._fx
			ymove = self.C * vertextmp._fy
		
			if (self.DEBUG): print "xmove =", xmove, "ymove =", ymove
			
			tmpposx = vertextmp._x + xmove
			tmpposy = vertextmp._y + ymove
			
			#//could gate here - i.e MAX_MOVE
			for j in xrange(len(self.NodeArray)):
				dist = row[j]
					
				if (dist == 0):
					continue
				
				adjvertextmp = self.NodeArray[j]
			
				if (self.DEBUG): print "adj X =", adjvertextmp._x, ",adj Y =", adjvertextmp._y
				if (self.DEBUG): print "tmp X =", tmpposx, "tmp Y =", tmpposy
				
				d = sqrt((adjvertextmp._x-tmpposx)*(adjvertextmp._x-tmpposx) + (adjvertextmp._y-tmpposy)*(adjvertextmp._y-tmpposy))
			
				if (self.DEBUG): print "d =", d, "- dist =", dist
				
				if ( d < dist and abs(d - dist) > 0.1):
					if (self.DEBUG): print "moving................"
	
					self.NodeArray[i]._x += xmove
					self.NodeArray[i]._y += ymove
					break
			
		
			if (self.DEBUG): print "---------------------------------------------------------f zeros for - ", i
			self.NodeArray[i]._fx = 0
			self.NodeArray[i]._fy = 0
	
	
	#//CORE FUNCTION - TIME CRITICAL
	def relax(self):
		for i in xrange(len(self.NodeArray)):
			if (self.DEBUG): print self.NodeArray[i]._x, self.NodeArray[i]._y, self.NodeArray[i]._fx, self.NodeArray[i]._fy
			row = self.DistanceMatrix[i]
		
			for j in xrange(len(self.NodeArray)):
				if i == j:
					continue
		
				dx = self.NodeArray[j]._x - self.NodeArray[i]._x
				dy = self.NodeArray[j]._y - self.NodeArray[i]._y
				d = sqrt(dx*dx + dy*dy)
				if(d < 0.000001):
					dx = 0.2
					d = sqrt(dx*dx + dy*dy)
				
				dist = row[j]
		
		
				if dist == 0:
				#//REPULSION
					repulsive_force = self.K * self.K / d
					if j != 0:
						self.NodeArray[j]._fx = self.NodeArray[j]._fx + repulsive_force * dx / d
						self.NodeArray[j]._fy = self.NodeArray[j]._fy + repulsive_force * dy / d
					if(i != 0):
						self.NodeArray[i]._fx = self.NodeArray[i]._fx - repulsive_force * dx / d
						self.NodeArray[i]._fy = self.NodeArray[i]._fy - repulsive_force * dy / d
				
				else:
				#//ATTRACTION
					repulsive_force = (dist - d) / (3*d)
					if j != 0:
						self.NodeArray[j]._fx = self.NodeArray[j]._fx + repulsive_force * dx
						self.NodeArray[j]._fy = self.NodeArray[j]._fy + repulsive_force * dy
					if i != 0:
						self.NodeArray[i]._fx = self.NodeArray[i]._fx - repulsive_force * dx
						self.NodeArray[i]._fy = self.NodeArray[i]._fy - repulsive_force * dy
	
	
if __name__ == '__main__':
	dmat  = [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	dmat += [[500,500,500,500,500,500,500]]
	
	dmat =  [[100, 500, 500, 500]]
	dmat += [[0, 100, 0, 0]]
	dmat += [[0, 0, 100, 0]]
	dmat += [[0, 0, 0, 100]]
	
	
	
	g = Graph(800,800)
	a = g.get_layout(dmat)
	
	for d in dmat:
		print d
	
	for k in a:
		print k._x, k._y, k._fx, k._fy