# Copyright (C) 2008, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


# Applications that use D-Bus to export a method to create a dictionary
import dbus
import dbus.service
import dbus.glib
import dbus.mainloop.glib
import gobject, traceback, sys, time

import os
from struct import pack, unpack
from random import random

#from cerebro.interface import *
from cerebro.common import *

def bin2mac(bin, all=False):
    if all:
        m = bin      # return all the bytes
        d = ''
    else:
        m = bin[-3:] # get the last three bytes
        d = ':'
        
    return d.join(["%02X" % ord(c) for c in m])

SERVICE = 'org.laptop.cerebro'

class Foo(object):
    def __init__(self):
        pass
    def one(self):
        print "one"
    def __getattr__(self, attr):
        print "intercepted"
        return lambda: getattr(self,name)
        
    
class CerebroInterface():
    def __init__(self, atype):
        self.ATYPE = atype
        self.aid = ''
        self.signup()
    def signup(self):
        SERVICE = 'org.laptop.cerebro'
        bus = dbus.SystemBus()
        iface = bus.get_object('org.laptop.cerebro', '/org/laptop/cerebro')
        #try:
        iface.connect_to_signal("on_node_arrival",   self._buddy_joined_cb)
        iface.connect_to_signal("on_node_info_arrival",self._buddy_info_cb)
        iface.connect_to_signal("on_node_departure", self._buddy_left_cb)
        iface.connect_to_signal("received_data",     self._received_cb)
        iface.connect_to_signal("on_request",        self._on_request_cb)
        #except:
            #pass
        
        self.register      = iface.get_dbus_method('register',     SERVICE)
        self.unregister    = iface.get_dbus_method('unregister',   SERVICE)
        self.push_data     = iface.get_dbus_method("push_data",    SERVICE)
        self.push_to       = iface.get_dbus_method("push_to",      SERVICE)
        self.show_net_tree = iface.get_dbus_method("show_net_tree",SERVICE)
        self.show_mst      = iface.get_dbus_method("show_mst",     SERVICE)
        self.show_dneighs  = iface.get_dbus_method("show_dneighs", SERVICE)
        self.send_command  = iface.get_dbus_method("exec_command", SERVICE)
        self.get_pers_info = iface.get_dbus_method("get_pers_info",SERVICE)
        self.set_pers_info = iface.get_dbus_method("set_pers_info",SERVICE)
        self.send_file     = iface.get_dbus_method("send_file",    SERVICE)
        self.get_aid       = iface.get_dbus_method("get_aid",      SERVICE)
        self.list_acts     = iface.get_dbus_method("list_acts",    SERVICE)
        self.get_stats     = iface.get_dbus_method("get_stats",    SERVICE)
        self.sendStats     = iface.get_dbus_method("sendStats",    SERVICE)
        self.load_file     = iface.get_dbus_method("load_file",    SERVICE)
            
    def printout(self, text):
        try:
            self.add_text(self.owner, text)
        except:
            print text
    
    def _received_cb(self, data, buddy, src, dst, port, aid, forme):
        data  = "".join(chr(b) for b in data)
        src   = "".join(chr(b) for b in src)
        dst   = "".join(chr(b) for b in dst)
        port  = "".join(chr(b) for b in port)
        print "aaaaaaaaaaa", data, buddy, src, dst, port, aid, forme
        return True
        try:
            nick = ''.join([chr(x) for x in buddy['nick']])
        except:
            nick = bin2mac(src)
            
        #self.printout("GOT from %s: port %s data: %s for: %s %d" % (nick, bin2mac(port), data, bin2mac(dst), forme))
                    
        if port == self.ATYPE:
            if aid == self.aid:
                self.printout("Chat: %s: %s" % (nick, data))
            elif data == 'stats':
                res = ""
                arrivals, overhead = self.get_stats()
                for a in arrivals:
                    res += "%s %s\n" % (a, arrivals[a])
                res += "\n"
                for line in overhead:
                    res += "%d %f\n" % (line[0], line[1])
                
                totneighs = len(arrivals)
                self.printout("total neighs: %d" % totneighs)
                t = totneighs * random()
                self.printout("will send back response in %f secs" % t)
                gobject.timeout_add(int(10*t), self.send_response, res, src)
            elif data[:4] == 'time':
                ltime = data[4:]
                try:
                    command = 'date -s %s' % ltime
                    self.printout("will attempt to run: %s" % command)
                    self.printout(os.popen( command ).read())
                except:
                    self.printout("******** failed to set time")
            elif aid == '0000000000' and forme:
                self.printout("extra from: %s %s" % (nick, data))
                if data[:3] == 'do ':
                    res = os.popen(data[3:]).read()
                    self.printout("will send back %s to %s" % (res, bin2mac(src)))
                    self.push_to(res, src, self.ATYPE)
            else:
                return True
        elif port == FILE_PORT:
            self.printout("File arrived: %s from: %s" % (data, nick))
        
    def send_response(self, data, dst):
        buddy = self.get_pers_info()
        nick  = ''.join([chr(x) for x in buddy['nick']])
        filename = nick
        f = open(filename, 'w')
        f.write(data)
        f.close()
        self.printout(self.send_file('', filename, dst))
        #self.push_to('', dst, self.ATYPE)
        
    def process_command(self, command):
        if command == "listactivities" or command == "la":
            lines = self.list_acts(0)
            for line in lines:
                res = ''
                for item in line:
                    res += str(item) + "\t"
                self.printout(res)
        if command == "tree":
            self.printout("Tree command temporarily disabled")
            return True
            for nick,path,dist,tot_dist,arrivals in self.show_net_tree():
                _path = ''
                for p in path:
                    _path += str(p) + ' '
                res = "%s %s %f %f %d" % (nick,_path,dist,tot_dist,arrivals)
                self.printout(res)
        if command == "mst":
            i = 0
            for buddy,path,dist,arrivals in self.show_mst():
                i += 1
                _path = ""
                for p in path:
                    _path += str(p) + " "
                res = "%d %s %s %f %d" % (i, buddy['nick'], _path, dist,
                    arrivals)
                self.printout(res)
        if command == "dn":
            i = 0
            for nick,dist,arrivals,first,last,period in self.show_dneighs():
                i += 1
                res = "%d %s %f %d %f %f %d" % (i, nick, dist, arrivals, first,
                    last, period)
                self.printout(res)
        if command == "sound":
            self.send_command('aplay /usr/share/sounds/alsa/Noise.wav')
        if command[:4] == "nick":
            newnick = command[5:]
            if newnick:
                myBuddyInfo = self.get_personal_info()
                myBuddyInfo['nick'] = newnick
                self.set_personal_info(myBuddyInfo)
        if command == "join":
            aid = command[5:]
            if not aid:
                aid = self.get_aid(self.ATYPE)
                if not aid:
                    self.printout("Please provide a valid aid to join")
                    return True
                
            self.printout("Will try to join activity ID: %s" % aid)
            try:
                self.aid = self.register(aid, self.ATYPE)
                res = "my aid is: %s" % self.aid
                self.printout(res)
            except dbus.DBusException:
                traceback.print_exc()
                sys.exit()
        if command == "leave":
            self.printout("Will leave all shared instancies")
            try:
                self.unregister(self.ATYPE)
            except dbus.DBusException:
                traceback.print_exc()
                sys.exit()
        if command == "share":
            self.printout("Will share current activity")
            try:
                self.aid = self.register('', self.ATYPE)
                res = "my aid is: %s" % self.aid
                self.printout(res)
            except dbus.DBusException:
                traceback.print_exc()
                sys.exit()
        if command[:4] == "send":
            try:
                dst = ''
                args = command[4:].split()
                if len(args) == 2:
                    (filename, dst) = args
                elif len(args) == 1:
                    filename = args[0]
                else:
                    self.printout("Example: /send FILENAME")
                    return 0
                self.printout(self.send_file(self.aid, filename, dst))
            except:
                traceback.print_exc()
        if command[:3] == "run":
            self.printout("Result of %s:" % command[4:])
            res = self.send_command(command[4:])
            self.printout(res)
        if command[:4] == "tell":
            try:
                args = command[4:].split()
                if len(args) > 1:
                    dst  = args[0]
                    data = command[6+len(dst):]
                else:
                    self.printout("Example: /tell DESTINATION DATA")
                    return 0
                res = self.push_to(data, dst, self.ATYPE)
                #self.printout("%d" % res)
            except:
                traceback.print_exc()
        if command[:4] == "time":
            a = time.localtime()
            ltime = "time %02d:%02d:%02d" % (a[3],a[4],a[5])
            try:
                self.printout("will push: %s" % ltime)
                res = self.push_to(ltime, 'all', self.ATYPE)
                #self.printout("%d" % res)
            except:
                traceback.print_exc()
        if command[:5] == "stats":
            try:
                self.printout("will ask for stats")
                self.sendStats()
                #self.printout("%d" % res)
            except:
                traceback.print_exc()

        # append an empty line
        self.printout(" ")
    

if __name__ == '__main__':
    cer = CerebroInterface()
    
    
