# Copyright (C) 2008, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sha, os
import time, sys
import demjson

from cerebro.common import *

COLOR_SIZE = 15
KEY_SIZE   = 1
IP_SIZE    = 4
IPV6_SIZE  = 16
SIGN_SIZE  = 20

class SAS: #Shared Activity Structure
    def __init__(self):
        self.idx = []
        self.nid = []
        self.aid = []
        self.typ = []
        self.inf = {}
    def new_buddy_info(self):
        buddy = {}
        buddy['nick']  = "<nonick>"
        buddy['color'] = " " * COLOR_SIZE
        buddy['pic']   = "a" * 2 * 1000
        buddy['key']   = "k" * KEY_SIZE
        buddy['ip']    = " " * IP_SIZE
        buddy['ipv6']  = " " * IPV6_SIZE
        return buddy
    def get_nick(self, mac):
        res = bin2mac(mac)
        for m in self.inf:
            if m == mac:
                res = self.inf[m]['nick']
        return res
    def get_mac(self, ident):
        for node in self.inf.items():
            if node[1]['nick'] == ident:
                return node[0]
        return ident
    def aid_by_nid(self, nid):
        res = []
        for i in xrange(len(self.nid)):
            if self.nid[i] == nid and (self.aid[i], self.typ[i]) not in res:
                res.append( (self.aid[i], self.typ[i]) )
        return res
    def nid_by_aid(self, aid):
        res = []
        for i in xrange(len(self.nid)):
            if self.aid[i] == aid and self.nid[i] not in res:
                res.append( self.nid[i] )
        return res
    def typ_by_aid(self, aid):
        for i in xrange(len(self.aid)):
            if self.aid[i] == aid:
                return self.typ[i]
        return ''
    def aid_by_typ(self, typ):
        for i in xrange(len(self.aid)):
            if self.typ[i] == typ:
                return self.aid[i]
        return ''
    def get_entry_bytype(self, typ):
        for i in xrange(len(self.nid)):
            if self.typ[i] == typ:
                return self.aid[i]
        return str(int(time.time()))
    def show_table(self):
        for i in xrange(len(self.nid)):
            print bin2mac(self.nid[i]), self.aid[i], bin2mac(self.typ[i])
    def add_entry(self, n, a, t):
        if n+a+t in self.idx:
            print "keeper.sas: entry already exists", n, a, t
            return True

        if len(n) != 6:
            print "PANIC: add_entry: got bad node id, expected 6 "+ \
                "bytes. Got:", n, len(n)
            sys.exit()
        if len(a) != 10:
            print "PANIC: add_entry: got bad activity id, expected 10 "+ \
                "bytes. Got:", a, len(a)
            sys.exit()
        if len(t) != 2:
            print "PANIC: add_entry: got bad activity type, expected 2" + \
                "bytes. Got:", t, len(t)
            sys.exit()
        self.idx.append(n+a+t)
        self.nid.append(n)
        self.aid.append(a)
        self.typ.append(t)
        if n not in self.inf:
            self.inf[n] = self.new_buddy_info()
    def del_entry(self, n, a, t):
        try:
            i = self.idx.index( n+a+t )
            del self.idx[i]
            del self.nid[i]
            del self.aid[i]
            del self.typ[i]
        except:
            print "SAS: del_entry: index not found for:", bin2mac(n), \
                bin2mac(a), bin2mac(t)
    def del_entry_bynid(self, n):
        for i in xrange(len(self.nid)-1,-1,-1):
            if self.nid[i] == n:
                del self.idx[i]
                del self.nid[i]
                del self.aid[i]
                del self.typ[i]
    def del_entry_bytype(self, t):
        for i in xrange(len(self.nid)-1,-1,-1):
            if self.typ[i] == t:
                del self.idx[i]
                del self.nid[i]
                del self.aid[i]
                del self.typ[i]
    def set_pers_info(self, nid, buddy):
        self.inf[nid] = buddy
    def __iter__(self):
        for i in xrange(len(self.idx)):
            yield self.nid[i], self.aid[i], self.typ[i]
    def clear(self):
        self.idx = []
        self.nid = []
        self.aid = []
        self.typ = []

class File:
    dic   = {}
    def __init__(self, name, fhash, data):
        self.dic['name' ] = name
        self.dic['fhash'] = fhash
        self.dic['data' ] = data
    def __str__(self):
        temp = self.dic.copy()
        del temp['data']
        temp['fhash'] = [ord(x) for x in self.dic['fhash']]
        return demjson.encode( temp, encoding='ascii' )
    def set_from_json(self, file_str):
        dic = demjson.decode( file_str, encoding='ascii' )
        dic['fhash'] = ''.join([chr(x) for x in dic['fhash']])
        self.dic = dic
        return dic
    def dictionary(self):
        return self.dic

class Cache:
    def __init__(self):
        self.hashes = ""
        self.files  = {}
    def add_hash(self, fhash, data, name):
        print "adding fhash:", bin2mac(fhash)
        self.hashes += fhash
        newfile = File(name, fhash, data)
        self.files[fhash] = newfile.dictionary()
    def in_cache(self, fhash):
        pos = self.hashes.find( fhash )
        return (pos > -1)
    def data_from_cache(self, fhash):
        if self.in_cache(fhash):
            return self.files[fhash]
        else:
            return {}
    def save_file(self, filename, data):
        file = open(FILEDIR + filename, 'wb+')
        file.write( data )
        file.close()

class Keeper(Cache):
    def __init__(self, nid, mp):
        Cache.__init__(self)
        self.nid = nid
        self.nid_size = len(nid)
        self.mp  = mp
        self.sas = SAS()
        self.buddy = {}
        #self.buddy['nick']   = "<nickname%d>" % int(time.time())
        self.buddy['nick' ] = NICKNAME # from common.py
        self.buddy['color'] = "#000000,#222222"
        pth = '/home/olpc/.sugar/default/config'
        if os.path.exists( pth ):
            self.buddy['nick' ] = os.popen("grep nickname "+pth+ \
                "|awk '{print $3}'").readline().strip()
            self.buddy['color'] = os.popen("grep color   "+pth+ \
                "|awk '{print $3}'").readline().strip()
        else:
            print "Probably not an XO. Setting nickname to",NICKNAME, \
                "color to default."

        self.buddy['sex' ] = "female"
        self.buddy['age' ] = "won't say!"
        self.buddy['in1' ] = "mp3s"
        self.buddy['in2' ] = "iPhone <$200"
        self.buddy['out1'] = "bicycle $70"
        self.buddy['out2'] = "laptop $500"
        self.buddy['ip'  ] = "X.X.X.X"
        self.buddy['pic' ] = MYPICTURE
        self.buddy['dorm'] = "East Campus"
        self.buddy['room'] = "12C"
        self.buddy['languages'] = "english, greek"


        self.share_file( 'pic', self.buddy['pic' ] )
        self.last_update = time.time()

    def share_file(self, file_key, filename):
        self.buddy[file_key] = filename
        print "will try to add key:", file_key
        print "file probably exists:", self.buddy[file_key]
        try:
            filename = self.buddy[file_key]
        except:
            print file_key,"is not in", self.buddy
            return 0
        try:
            data = open(FILEDIR + filename, 'r').read()
        except:
            print "File:", FILEDIR + filename, "does NOT exist!"
            return 0

        shahash = sha.new( data )
        fhash = shahash.digest()
        print "File:", self.buddy
        print "data:", len(data)
        self.add_hash( fhash, data, filename )
        return str(File(filename, fhash, data))

    def pack_status_info(self, nid):
        if nid != self.nid and not nid in self.sas.nid:
            # return if I have no info about requested nid
            return ""

        _aid = []
        _typ = []
        for i in xrange(len(self.sas.nid)):
            if self.sas.nid[i] == nid:
                #res += self.sas.aid[i] + self.sas.typ[i]
                _aid.append(self.sas.aid[i])
                _typ.append(self.sas.typ[i])

        #shahash = sha.new( res + self.buddy['key'] )
        #signature = shahash.digest()
        #res = self.nid + signature + res

        self.buddy['__nid__'] = [ord(x) for x in nid]
        self.buddy['__aid__'] = _aid
        self.buddy['__typ__'] = _typ

        if nid == self.nid:
            buddy = self.buddy
        else:
            buddy = self.sas.inf[nid]

        print
        print "will encode:", buddy
        print
        res = demjson.encode( buddy, encoding='ascii' )

        return str(res)
    
    def unpack_status_info(self, src, fhash, data):
        self.add_hash( fhash, data, bin2mac(src) )

        b = demjson.decode( data, encoding='ascii' )

        buddy = {}
        for x in b:
            val = b[x]
            if type(val) == unicode:
                buddy[str(x)] = str(val)
            elif type(val) == list:
                newval = []
                for v in val:
                    if type(v) == unicode:
                        newval.append(str(v))
                    else:
                        newval.append(v)
                buddy[str(x)] = newval
            else:
                buddy[str(x)] = val

        nid = ''.join([chr(x) for x in buddy['__nid__']])

        self.sas.inf[nid] = buddy

        if nid in self.sas.inf:
            print "update userinfo: ", bin2mac(nid), "is", buddy['nick']
            return False
        else:
            print "new userinfo: ", bin2mac(nid), "is", buddy['nick']
            self.sas.nid += len(buddy['__aid__']) * [nid]
            self.sas.aid += buddy['__aid__']
            self.sas.typ += buddy['__typ__']
            return True

    def my_status_hash(self):
        shahash = sha.new( self.my_status_info() )
        return    shahash.digest()

    def set_pers_info(self, nid, buddy):
        self.sas.set_pers_info(nid, buddy)
        self.last_update = time.time()
        #if nid not in self.inf:
            #self.inf[nid] = buddy
        #for prop in buddy:
            #if a in dir(self.inf[nid]): # if this is a valid property of the buddy dictionary
                #self.inf[nid].__dict__[a] = args[a]



