# Copyright (C) 2008, Polychronis Ypodimatopoulos
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys, time, select, os, sha, threading
from string import upper
from struct import pack, unpack
from math   import ceil
from socket import *
from random import random
import json
import gobject

from cerebro.common import *


#def bin2mac(bin):
    #char = ''
    #if len(bin) == 6:
        #char = ':'
    #return char.join(["%02X" % ord(c) for c in bin])

def pout(*args):
    port = args[0]
    prefix = "test %f:" % time.time()
    

    debug_level = args[0]

    # debug_level of 0 means ALWAYS print
    if port == '\x02\x00' or port == "":
        if len(args):
            print prefix+":", ' '.join([str(el) for el in args])

#START_TIME = int(time.time()/1000)*1000 # make sure that all machines start within a 1000 sec frame

#try:
    #import pcapy
#except:
    #pout( "pcapy module not available")
###################################################################################

class FileStats:
    def __init__(self, nid_size):
        self.nid_size      = nid_size
        self.receptions    = {}
        self.transmissions = {}
        self.START_TIME    = 0
    
    def export_to_json(self):
        r = ""
        for fhash in self.receptions:
            r += self.receptions[fhash]

        #print "dictionary:"
        #print self.receptions
        #print "r:"
        #print r
        t = ""#json.write( self.transmissions )
        return (r, t)
    
    #######
    def log_reception_start(self, receiver):
        if receiver.port != '\x02\x00':
            return -1
        #self.receptions[ receiver.fhash ] = [ receiver.total_pieces, [] ]
        self.receptions[ receiver.fhash ] = "%s, %d, 0, 0\n" % (bin2mac(receiver.fhash), receiver.total_pieces)
        
    def log_reception_timepoint(self, receiver):
        if receiver.port != '\x02\x00':
            return -1
        tp = self.receptions[ receiver.fhash ] # pointer to timepoints
        #tp.append( [time.time(), sum(receiver.acked), receiver.total_opportunistic, receiver.bytes_recvd, receiver.origins ] )
        if not self.START_TIME: self.START_TIME = time.time()
        self.receptions[ receiver.fhash ] += "%f, %d, %d, %d\n" % (time.time()-self.START_TIME, sum(receiver.acked), receiver.total_opportunistic, receiver.bytes_recvd)
        
        
    #######
    def log_transmission_start(self, sender):
        if sender.port != '\x02\x00':
            return -1
        self.transmissions[ sender.dst + sender.fhash ] = [ time.time() ]	# 0
        
    def log_transmission_end(self, sender, successful):
        if sender.port != '\x02\x00':
            return -1
        pointer = self.transmissions[ sender.dst + sender.fhash ]
        pointer.append( sender.rtt )										# 1
        pointer.append( sender.burst_size )									# 2
        pointer.append( sender.loss_r )										# 3
        pointer.append( sender.packet_size )								# 4
        pointer.append( sender.port )										# 5
        pointer.append( sender.outstanding )								# 6
        pointer.append( sender.MAX_RETRIES )								# 7
        pointer.append( sender.next_pno )									# 8
        pointer.append( sender.total_sent )									# 9
        pointer.append( sender.total_pieces )								# 10
        #pointer.append( sender.usebroadcast )								# 11
        pointer.append( sender.max_acked )									# 12
        pointer.append( successful )										# 13
        pointer.append( time.time() )										# 14
        pointer.append( sender.total_size )									# 15
        
        
    def export_stats(self):
        print "TXs:"
        grand_total = 0
        total_txs = 0
        min_start = 0
        max_end = 0
        for s in self.transmissions:
            p = self.transmissions[s]
            sent = p[9]
            
            start_time = p[0]
            end_time = p[14]
            
            if end_time > max_end:
                max_end = end_time
            if start_time < min_start or min_start == 0:
                min_start = start_time
            
            total_time = end_time - start_time
            filesize = p[15]
            speed = 1.0*filesize/(total_time*1024)
            print "fhash", bin2mac(s[self.nid_size:]), "to:", bin2mac(s[:self.nid_size])
            print "TOTAL SENT:  %d bytes in %f seconds (Goodput: %f KB/s)" % (sent, total_time, speed)
            print "Original file size: %d bytes" % filesize
            stretch = 1.0*sent/filesize
            print "STRETCH (including headers): %f" % stretch
            print
            grand_total += sent
            total_txs += 1
            
        print min_start
        print max_end
        tot_time = max_end-min_start
        print tot_time
        if tot_time == 0:
            speed = -1
        else:
            speed = 1.0 * grand_total / (tot_time*1024)
        
        print "I sent a total of", grand_total,"bytes in", total_txs,"transmissions in", tot_time,"secs"
        print "SPEED:", speed, "KBytes/sec"
        print
        print
            
        print "RXs:"
        for r in self.receptions:
            p = self.receptions[r]
            tp = p[1]
            last_tp = tp[-1]
            first_tp = tp[0]
            
            start_time = first_tp[0]
            end_time = last_tp[0]
            total_time = end_time - start_time
            
            total_recvd = last_tp[1]
            total_oppor = last_tp[2]
            filesize = last_tp[3]
            if total_time > 0:
                speed = 1.0*filesize/(total_time*1024)
            else:
                speed = -1
            origins = last_tp[4]
            
            print "fhash:", bin2mac(r)
            print "total pieces       :", p[0]
            print "time points        :", len(tp)
            print "total received     :", total_recvd, "(completed:",100.0*total_recvd/p[0],"%)"
            print "total opportunistic:", total_oppor, "(",100.0*total_oppor/total_recvd,"%)"
            print "start time         :", start_time
            print "end time           :", end_time
            print "total time         :", total_time
            print "filesize           :", filesize
            print "goodput            :", speed, "KBytes/sec"
            print "origins            :", origins
            print
            

        """
        fileHandle = open( 'RTT-cum-retx-0.0%d.dat' % (self.rtt*1000), 'w' )
        for line in self.cumretx:
            fileHandle.write( line )
        fileHandle.close()
        fileHandle = open( 'RTT-pno-retx-0.0%d.dat' % (self.rtt*1000), 'w' )
        for i in xrange(self.tot_pieces):
            fileHandle.write( "%d %d\n" % (i, self.retx[i]) )
        fileHandle.close()

        fileHandle = open( 'globalStats.dat', 'a' )
        fileHandle.write( "%f %f %f %f %f\n" % (self.filesize, total_time, speed, speed/stretch, stretch) )
        fileHandle.close()
        """
###################################################################################

class NetworkInfo:
    def __init__(self, myaddr, send_fracket):
        self.myaddr      = myaddr
        self.send_fracket = send_fracket
        
    def send_ack_frame(self, load, port, dst_addr="", src_addr=""):
        return self.send_frame(ACK_FRAME, load, dst_addr, src_addr, 1, port)
        
    def send_data_frame(self, load, port, dst_addr="", src_addr="", burst=1):
        return self.send_frame(FILE_FRAME, load, dst_addr, src_addr, burst, port)
    
    def send_req_frame(self, load, dst_addr, port, src_addr=""):
        return self.send_frame(REQ_FILE_FRAME, load, dst_addr, src_addr, 1, port)

    def send_frame(self, ftype, load, dst_addr, src_addr, burst, port):
        output = ftype
        try:
            output += pack("H", burst)
        except:
            print "WHY ????? burst:", burst
            sys.exit()
        output += port
        output += pack("H", len(load))
        output += load
        return self.send_fracket( dst_addr, src_addr, ftype, output )

###################################################################################

class Sender(threading.Thread):
    def __init__(self, net, dst, data, fhash, port, stats, completion_cb, rtt=0.05, burst_size=50, loss_r=0, packet_size=1450):
        pout( port, "instantiating Sender for fhash", bin2mac(fhash), "to", bin2mac(dst))
        self.net          = net
        
        self.fhash        = fhash
        self.completion_cb= completion_cb
        
        self.rtt          = rtt
        self.burst_size   = burst_size
        self.loss_r       = loss_r
        self.packet_size  = packet_size
        
        self.dst          = dst
        self.port         = port
        self.got_ack      = 0
        self.outstanding  = []
        self.MAX_RETRIES  = 20
        self.next_pno     = 0
        self.total_sent   = 0
        self.total_pieces = int(ceil( 1.0 * len(data) / self.packet_size ))
        self.total_size   = len(data)
        
        self.data         = self.prepare_transmission(data, dst, fhash)
        
        self.max_acked    = -1
        self.lock         = threading.Lock()
        threading.Thread.__init__(self)
        self.stats        = stats
        self.stats.log_transmission_start( self )
        
        #print "teleporter: started SENDER, fhash:", bin2mac(fhash), "port", bin2mac(port), "load", bin2mac(data), "totalpno", self.total_pieces
        
    def process_acks(self, last, nacks, src):
        #print "    GOT ACK     sending to:",bin2mac(self.dst),"         last:", last, "       from:", bin2mac(src)
        
        # nacks is a series of bytes. Every 2 of them is an integer (0-65535)
        self.outstanding = []
        for x in xrange(0,len(nacks),2):
            tmp = unpack("H", nacks[x:x+2])[0]
            self.outstanding.append( tmp )
        #print "outstanding1:", self.outstanding
            
        # last_pno_acked is the maximum index of the piece that the receiver has gotten
        # so far. All pieces following this piece up till the last piece that the sender
        # sent (self.next_pno) must be retransmitted. Since last_pno_acked was actually
        # received, we start with the next one and place them in self.outstanding.
        last_pno_acked = unpack("H", last)[0]
        for x in xrange(last_pno_acked + 1, self.next_pno, 1):
            self.outstanding.append( x )
        
        if self.max_acked < last_pno_acked: self.max_acked = last_pno_acked
        #else: print "self.max_acked:", self.max_acked, " now got:", last_pno_acked
        
        if self.next_pno < self.max_acked:
            #print "SENDER: Receiver acked a piece in the future."
            self.next_pno = self.max_acked + 1
            
        #self.stats.log_acks( len(self.outstanding) )
        #print "outstanding2:", self.outstanding
        #print "last_pno_acked:", last_pno_acked
        
        self.got_ack = 1
        
        
    def prepare_transmission(self, data, dst, fhash):
        result = []
        i = 0
        _total_pieces = pack("H", self.total_pieces )
        for i in xrange(self.total_pieces):
            packet = fhash + pack("H", i) + _total_pieces + data[(i*self.packet_size):(i+1)*self.packet_size]
            result.append( packet )
            i += 1

        return result
    
    
    def run(self):
        pout( self.port, "TOTAL pieces TO SEND:", self.total_pieces)
        
        #if not (self.next_pno < self.total_pieces or self.outstanding):
            #print "WOW: Sender finished before it even started: to:",bin2mac(self.dst)
        counter = 0
        while self.next_pno < self.total_pieces or self.outstanding:
            #i += 1
            self.send_burst()
            counter = 0
            while counter < self.MAX_RETRIES and not self.got_ack:
                self.send_single_packet()
                time.sleep( self.rtt ) # retry timeout
                counter += 1
                
            if counter == self.MAX_RETRIES:
                pout( self.port, "Giving up")
                self.completion_cb(self.dst, self.fhash, self.port, False, counter)
                self.stats.log_transmission_end( self, False )
                return -1
            
            self.lock.acquire()
            self.got_ack = 0
            self.lock.release()
            
        #if i == 0:
            #print "NO bursts sent at all!"
            
        self.completion_cb(self.dst, self.fhash, self.port, True, counter)
        self.stats.log_transmission_end( self, True )
        
        #print "sender dying:", self
        
    def send_burst(self):
        burst_size = self.burst_size
        # burst_size must not exceed the number of packets that will be sent
        if self.total_pieces - self.next_pno < burst_size:
            burst_size = self.total_pieces - self.next_pno
        
        # send the outstanding frames first
        for p in self.outstanding:
            self.send_packet( p, burst_size, 1 )
                
        # send a burst, plus self.loss_r of the burst_size
        upper_limit = int(self.next_pno + (1+self.loss_r)*burst_size - len(self.outstanding))
        if upper_limit > self.total_pieces:
            upper_limit = self.total_pieces
        
        pout( self.port, "Will try a burst from %d to %d (burst:%d) (total:%d) (outstanding:%d)" % (self.next_pno, upper_limit - 1, burst_size, self.total_pieces, len(self.outstanding)))
            
        for p in xrange(self.next_pno, upper_limit, 1):
            if p > self.next_pno + burst_size - 1:
                pout( self.port, "Sending packet:", p)
                burst_size = 1
                
            self.send_packet( p, burst_size, 2 )
        
        if self.next_pno > upper_limit:
            pass
            #print "SENDER: Receiver acked faster than I expected. I will keep up."
        else:
            self.next_pno = upper_limit
            

    def send_single_packet(self):
        if self.next_pno < self.total_pieces:
            self.send_packet(self.next_pno, 1, 3)
            #print "RETRY: pno: %d / %d" % (self.next_pno, self.total_pieces)
            self.next_pno += 1
        else:
            self.send_packet(self.total_pieces - 1, 1, 4)
            #print "RETRY: pno: %d / %d" % (self.total_pieces - 1, self.total_pieces)
            
        
    def send_packet(self, index, burst_size, where):
        #dst_address_to_use = self.dst
        #if self.usebroadcast:
            #dst_address_to_use = self.BROADCAST
        sent = self.net.send_data_frame(self.data[index], self.port, burst=burst_size, dst_addr=self.dst)
        self.total_sent += sent
        if where > 4:
            pout(self.port, "       ",where,"     to:",bin2mac(self.dst),"             sent/tot_sent:", sent, self.total_sent)
        return sent
            
            
###################################################################################


class Requestor(threading.Thread):
    def __init__(self, net, dst, request, port, completion_cb, rtt=0.05, packet_size=1450):
        pout( port, "instantiating Requestor to", bin2mac(dst))
        self.net          = net
        
        self.completion_cb= completion_cb
        self.got_data_frame = 0
        self.rtt          = rtt
        self.packet_size  = packet_size
        
        self.dst          = dst
        self.port         = port
        self.MAX_RETRIES  = 5
        self.request      = request
        
        threading.Thread.__init__(self)
    
    def run(self):
        retries = 0
        while retries < self.MAX_RETRIES and not self.got_data_frame:
            pout( self.port, "Requestor going", retries, "for:", bin2mac(self.dst))
            sleeptime = self.rtt*random()*100
            print "Requestor going", retries, "for:", bin2mac(self.dst), "will sleep for", sleeptime
            self.net.send_req_frame(self.request, self.dst, self.port)
            
            time.sleep( sleeptime ) # retry timeout
            retries += 1
        
        outcome = True
        if retries == self.MAX_RETRIES:
            outcome = False
            pout( self.port, "Requestor: Giving up")

        self.completion_cb(self.dst, self.port, self.request, outcome, retries)

###################################################################################

class Receiver:
    def __init__(self, network, fhash, port, ptotal, packet_size):
        pout( port, "instantiating Receiver for fhash", bin2mac(fhash))
        self.network = network
        self.fhash = fhash
        self.port = port
        self.total_pieces = ptotal
        self.recvd_pieces = 0 # this only holds the total recvd pieces during the current burst
        self.max_pno = -1
        self.packet_size = packet_size
        self.target = [] # target (probably a bad name) stores all data received so far
        self.bytes_recvd = 0 # total unique bytes received so far
        self.acked = []
        self.origins = {}
        self.total_opportunistic = 0
        self.request = "" # original request that triggered this reception (can be empty)
        
        for k in xrange(ptotal):
            self.acked.append( 0 )
            self.target.append( "+" )
        #self.stats = FileStats(fhash, len(data), self.total_pieces, RTT)
        
    
    def send_ack(self, dst):
        ack_vector = pack("H", self.max_pno)
        
        #print "sending back ack:", self.max_pno
        for p in xrange(self.max_pno):
            if not self.acked[p]:
                ack_vector += pack("H", p)
                
        self.send_packet( dst, ack_vector )
    
        
    def process_data_frame(self, src, dst, burst_size, load):
        pno   = unpack("H", load[20:22])[0]
        piece = load[24:]
        # if this is the first piece, get the original request from it
        if pno == 0:
            #print "     Receiver: got piece no 1 for:", bin2mac(self.fhash), "from", bin2mac(src)
            length = unpack("H", piece[:2])[0] # the first two bytes contain the length of the request
            self.request = piece[2:2+length]
            piece = piece[2+length:] # the actual data is what follows
        
        # Increase received pieces
        self.recvd_pieces += 1
        
        if self.acked[pno] == 0:
            if src in self.origins:
                self.origins[src] += 1
            else:
                self.origins[src]  = 1
        
        # if this piece was not meant for me and I did not have it already
        if self.network.myaddr != dst and self.acked[pno] == 0:
            self.total_opportunistic += 1
            if self.total_opportunistic % 20 == 0:
                pout( "", "Total opportunistic pieces: %d/%d (%f %%)" % ( self.total_opportunistic, self.total_pieces, 100.0*self.total_opportunistic/self.total_pieces ) )
                
        previous = self.acked[pno]
        self.acked[pno] = 1
        
        # update the max pno received so far
        if pno > self.max_pno: self.max_pno = pno
        
        # If we reached the burst size or all pieces are received
        s = sum(self.acked)
        if self.recvd_pieces >= burst_size or s == self.total_pieces:
            if self.network.myaddr == dst:
                self.send_ack( src )
                self.recvd_pieces = 0
        
        # check if we already received this
        if previous:
            #print "Already have piece:", pno
            return (-1, self.request)
        else:
            self.target[pno] = piece
        
        self.bytes_recvd += len(piece)
        
        if s == self.total_pieces:
            return (2, self.request)
        else:
            #print "Progress: %d/%d" % (s, self.total_pieces)
            return (1, self.request)
        
    def send_packet(self, dest, data):
        sent = self.network.send_ack_frame(self.fhash + data[:self.packet_size], self.port, dst_addr=dest)
        data = data[self.packet_size:]
        
        while data:
            sent += self.network.send_ack_frame(self.fhash + data[:self.packet_size], self.port, dst_addr=dest)
            data = data[self.packet_size:]

        return sent
###################################################################################

class Teleporter(gobject.GObject):

    __gsignals__ = {
        'requestor-finished' : (gobject.SIGNAL_ACTION, gobject.TYPE_NONE,
            (gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_STRING,
             gobject.TYPE_BOOLEAN, gobject.TYPE_INT)),
        'sender-finished' : (gobject.SIGNAL_ACTION, gobject.TYPE_NONE,
            (gobject.TYPE_STRING,  gobject.TYPE_STRING, gobject.TYPE_INT,
             gobject.TYPE_BOOLEAN, gobject.TYPE_INT))
    }

    def __init__(self, myaddr, send_fracket):
        self.FTYPE			= 0
        self.BURST_SIZ		= 1
        self.PORT			= 2
        self.LOAD			= 3

        self.receivers		= {}
        self.senders		= {}

        self.outstanding_requests   = {}
        self.outstanding_receptions = {}

        self.PACKET_SIZE	= 1450
        self.RTT			= 0.1
        self.BURST_SIZE		= 50
        self.LOSS_R			= 0.04

        self.nid_size		= len(myaddr)
        self.lock			= threading.Lock()
        self.net			= NetworkInfo( myaddr, send_fracket )
        self.neteng			= None
        self.handle_new_data= None

        self.mcast_destinations = []
        self.mcast_data = ""

        self.stats			= FileStats(self.nid_size)
        #pout( "", "\n"*5 )
        #pout( "", "Starting Teleporter")
        #pout( "", "Socket type       :", socket_type)
        #pout( "", "Wireless interface:", self.net.wface)

    def process_data(self, src, dst, raw_data):
        fr = []
        # minus 14
        fr.append( raw_data[0] )
        fr.append( raw_data[1:3] )
        fr.append( raw_data[3:5] )
        loadlen   = unpack("H", raw_data[5:7])[0]
        fr.append( raw_data[7:7+loadlen] )

        if src == self.net.myaddr:
            return 0

        if fr[self.FTYPE] == ACK_FRAME:
            port = fr[self.PORT]
            if dst == self.net.myaddr:
                self.lock.acquire()
                fhash = fr[self.LOAD][:20]
                key = src + fhash + port
                try:
                    self.senders[ key ].process_acks( fr[self.LOAD][20:22],
                        fr[self.LOAD][22:], src )
                except KeyError:
                    #pout(port, " Received an ack for non-existent sender", bin2mac(src), bin2mac(fhash), bin2mac(port))
                    pass
                self.lock.release()
            
        elif fr[self.FTYPE] == FILE_FRAME:
            
            load = fr[self.LOAD]
            
            fhash = load[:20]
            port  = fr[self.PORT]
                
            # if fhash already exists then serve from cache
            if self.neteng.in_cache( fhash ):
                pno, ptotal	= unpack( "HH", load[20:24] )
                if dst == self.net.myaddr:
                    ack = load[:20] + pack("H", ptotal)
                    sent = self.net.send_ack_frame( ack, port, src )
                piece = load[24:]
                
                # if this is the first piece, get the original request from it
                if pno == 0:
                    #print "     Receiver: got piece no 1 for:", bin2mac(self.fhash), "from", bin2mac(src)
                    length = unpack("H", piece[:2])[0] # the first two bytes contain the length of the request
                    request = piece[2:2+length]
                    piece = piece[2+length:] # the actual data is what follows
                    key = src + port + request
                    if key in self.outstanding_requests:
                        self.outstanding_requests[ key ].got_data_frame = 1
                        self.neteng.handle_new_data_from_cache(fhash, src, dst, port)
                return 0
            ###### end serve from cache
                
            if not fhash in self.receivers:
                ptotal = unpack("H", load[22:24])[0]
                self.receivers[fhash] = Receiver( self.net, fhash, port, ptotal, self.PACKET_SIZE )
                pout( port, "started new RECEIVER for:", bin2mac(fhash), ptotal)
                
                if fhash in self.outstanding_receptions:
                    (tstart, retries) = self.outstanding_receptions[ fhash ]
                    self.outstanding_receptions[ fhash ] = (tstart, retries + 1)
                else:
                    self.outstanding_receptions[ fhash ] = (time.time(), 1)
                    
                self.stats.log_reception_start( self.receivers[fhash] )
                
                
            burst_size = unpack("H", fr[self.BURST_SIZ])[0]
            (verdict, request) = self.receivers[fhash].process_data_frame(src, dst, burst_size, load)
            
            
            # keep track of outstanding requests
            if verdict > 0:
                # log receptions
                self.stats.log_reception_timepoint( self.receivers[fhash] )
                
                # I just got the first packet and I know what request this corresponds to
                # if I actually requested this myself, then eliminate the outstanding request
                key = src + port + request
                
                if key in self.outstanding_requests:
                    self.outstanding_requests[ key ].got_data_frame = 1
                
            # last piece missing was just received
            if verdict == 2:
                if port == FILE_PORT:
                    pout(port, "It looks like we're done!", bin2mac(key))
                data = ""
                for p in self.receivers[fhash].target:
                    data += p
                
                newhash = sha.new( data )
                newhash = newhash.digest()

                if fhash == newhash:
                    #print "teleporter calling handle_new_data", len(data)
                    self.handle_new_data(data, fhash, src, dst, port)
                    del self.receivers[fhash]
                    #print "receiver deleted", len(self.receivers)
                else:
                    print "OOOPS! file is different from original:"
                    print "original hash:", bin2mac(fhash)
                    print "new hash     :", bin2mac(newhash)
                    sys.exit()
                
                # if the reception is finished, then eliminate the outstanding reception
                if key in self.outstanding_receptions:
                    (tstart, retries) = self.outstanding_receptions[ key ]
                    total_time = time.time() - tstart
                    pout(port, "deleting outstanding reception for:", bin2mac(key))
                    del self.outstanding_receptions[ key ]
            
        elif fr[self.FTYPE] == REQ_FILE_FRAME:
            port = fr[self.PORT]
            load  = fr[self.LOAD]
            
            pout(port, "IT'S A REQ from", bin2mac(src))
                
            # take care of statistics
            #self.stats.log_response_time
            
            # if the port is registered with keeper, get the data to be sent
            contents = self.neteng.handle_new_req( src, port, load )
            if contents:
                shahash = sha.new( contents )
                fhash   = shahash.digest()
                pout(port, "Computed hash: %s len of data: %d for: %s" % (bin2mac(fhash), len(contents), bin2mac(src)))
                
                # check to see if there is already a sender for the specific fhash
                senders_copy = self.senders.copy()
                for key in senders_copy:
                    if fhash + port == key[self.nid_size:]:
                        pout( port, "Hash", bin2mac(fhash), "is already being sent to", bin2mac(key[:self.nid_size]),"skipping request from", bin2mac(src) )
                        return 0
                
                #include the original request at the start of the contents
                contents = pack("H", len(load)) + load + contents
                
                # instantiate a sender
                sender = Sender( self.net, src, contents, fhash, port, self.stats, self.sender_cb )
                sender.rtt         = self.RTT
                sender.burst_size  = self.BURST_SIZE
                sender.loss_r      = self.LOSS_R
                sender.packet_size = self.PACKET_SIZE
                
                self.senders[src + fhash + port] = sender
                sent = sender.start()
                print "sender started", len(self.senders)
            else:
                print "Teleporter: Got req from %s: No app listenting on" + \
                    "port: %d %d, or request generated empty response" \
                    % ( bin2mac(src), ord(port[0]), ord(port[1]) )
            
        elif fr[self.FTYPE] == "8":
            print "Reporting status:"
            self.stats.export_stats()
            #for r in self.receivers:
                #print "total pieces       : %d out of %d" % (sum( self.receivers[r].acked ), self.receivers[r].total_pieces)
                #print "total opportunistic:", self.receivers[r].total_opportunistic
                #for o in self.receivers[r].origins:
                    #print "Origin %s: %d pieces" % (bin2mac(o), self.receivers[r].origins[o])
            
        else:
            print "\tI got a weird packet of type: %d" % ord( fr[self.FTYPE] )
            sys.exit()
        
        del fr
        
    
    def requestor_cb(self, dest, port, request, outcome, retries):
        #print "requestor completed with:", outcome,"req:", request, "to", bin2mac(dest), "port", port, "retries:", retries
        if not outcome:
            print "failed to get request through to", bin2mac(dest), "on port", unpack("H", port)[0]
            
        pout(port,  "Deleting requestor for:", request, "to:", bin2mac(dest), "result:", outcome)
        del self.outstanding_requests[dest + port + request]
        
    def requestor_multicast_cb(self, dest, port, request, outcome, retries):
        if dest:
            print "PREVIOUS ROUND FINISHED with:", outcome, "to dest", bin2mac(dest), "retries", retries
            
        if self.mcast_destinations:
            next = self.mcast_destinations.pop()
            print "NOW SENDING to:", bin2mac(next)
            self.send_request(next, port, self.mcast_data, callback=self.requestor_multicast_cb)
        
    def send_request(self, addr, port, request, callback=None):
        #if not addr: addr=self.net.BROADCAST
        if not callback: callback = self.requestor_cb
        
        # keep track of outstanding requests
        key = addr + port + request
        if key in self.outstanding_requests:
            print "teleporter: there is a pending request already"
            return True
        
        pout( port, "sending request", bin2mac(request), bin2mac(addr), port)
        
        requestor = Requestor(self.net, addr, request, port, callback)
        requestor.rtt         = self.RTT
        requestor.packet_size = self.PACKET_SIZE
        
        pout(port, "created outstanding_requests for", bin2mac(addr), bin2mac(port), request)
        self.outstanding_requests[key] = requestor
        sent = requestor.start()
        print "requestor started", len(self.outstanding_requests)
        
    def send_multicast_request(self, destinations, port, request):
        print "Teleporter: got MULTICAST request"
        self.mcast_destinations = destinations
        self.mcast_data = request
        self.requestor_multicast_cb("", port, "", False, -1)
        
    def sender_cb(self, dest, fhash, port, success, retries):
        if not success:
            print "failed to send to", bin2mac(dest), "on port", unpack("H", port)[0]
            
        pout(port,  "Deleting sender for:", bin2mac(fhash), "to:", bin2mac(dest), "result:", success)
        
        del self.senders[dest + fhash + port]
        print "sender deleted", len(self.senders)
    
    
    def sender_multicast_cb(self, dest, fhash, port, success, retries):
        if dest:
            print "PREVIOUS ROUND FINISHED with:", success, "to dest", bin2mac(dest), "retries", retries
            
        if self.mcast_destinations:
            next = self.mcast_destinations.pop()
            print "NOW SENDING to:", bin2mac(next)
            self.push_data(next, self.mcast_data, port, callback=self.sender_multicast_cb)

        
    def push_multicast_data(self, destinations, port, data):
        self.mcast_destinations = destinations
        self.mcast_data = data
        self.sender_multicast_cb("", "", port, False, -1)
            
                    
    def push_data(self, dest, data, port, callback=None):
        if not callback: callback = self.sender_cb
        
        shahash = sha.new( data )
        fhash   = shahash.digest()
        request = "" # this is a PUSH, so the request is empty
        data = pack("H", len(request)) + request + data
        
        #print "teleporter: will push:", bin2mac(dest), data, bin2mac(fhash), port
        
        sender = Sender( self.net, dest, data, fhash, port, self.stats, callback )
        sender.rtt         = self.RTT
        sender.burst_size  = self.BURST_SIZE
        sender.loss_r      = self.LOSS_R
        sender.packet_size = self.PACKET_SIZE
        
        self.senders[dest + fhash + port] = sender
        sent = sender.start()
        #print "sender started", len(self.senders)
        
    
    def publish_update(self, data): # to bcast address
        pass

    def process_requests(self):
        pass
        
    def process_receptions(self):
        pass
    

