# !/usr/bin/env python

# Applications that use D-Bus to export a method to create a dictionary
import dbus
import dbus.service
import dbus.glib
import dbus.mainloop.glib
import gobject, traceback, sys, time
from struct import pack, unpack

def bin2mac(bin, all=False):
	if all:
		m = bin      # return all the bytes
		d = ''
	else:
		m = bin[-3:] # get the last three bytes
		d = ':'
		
	return d.join(["%02X" % ord(c) for c in m])

#ATYPE = '\x0a\x00'
ATYPE = '00'
SERVICE = 'org.olpc.cerebro'
OBJECT_PATH = '/org/olpc/cerebro/chat123'
INTERFACE = 'org.olpc.cerebro.chat123'

class Chat(dbus.service.Object):

	def __init__(self):

		bus = dbus.SystemBus()
		interface = bus.get_object('org.olpc.cerebro', '/org/olpc/cerebro')
		interface.connect_to_signal("on_node_arrival", self.onNodeArrival)
		interface.connect_to_signal("received_data", self.receive_chat_text)
		
		self.push_data = interface.get_dbus_method("push_data", 'org.olpc.cerebro')
		
		self.buddyinfo = []
		self.last = {}
		
		try:
			register = interface.get_dbus_method('register', 'org.olpc.cerebro')
			self.aid = register(ATYPE, SERVICE, '/org/olpc/cerebro', INTERFACE, OBJECT_PATH)
			print "my aid is:", self.aid, len(self.aid), bin2mac(self.aid,1)
		except dbus.DBusException:
			traceback.print_exc()
			sys.exit()
		
	def onNodeArrival(self, node_arr):
		for n in node_arr:
			s = "".join(chr(b) for b in n)
			#s = "".join(ord(b) for b in n)
			print "CHAT: Node arrived", bin2mac(s)
			print "sending to", bin2mac(s)
			self.send_chat_text('hello world')

	def send_chat_text(self, data):
		sequence = pack('q', 1000 * time.time()) # 8 bytes
		print "chat: sending", data, len(self.aid)
		data = sequence + data
		self.push_data(self.aid, data, ATYPE)
		
	def receive_chat_text(self, data, buddy, src, port, aid):
		data  = "".join(chr(b) for b in data)
		#fhash = "".join(chr(b) for b in fhash)
		src   = "".join(chr(b) for b in src)
		port  = "".join(chr(b) for b in port)
		
		if port != ATYPE or aid != self.aid:
			return True
		
		#buddy = bin2mac(src)
		nick = ''.join([chr(x) for x in buddy['nick']])
		
		sequence = data[:8]
		text     = data[8:]
		sequence = unpack('q', sequence)[0]
		if not src in self.last:
			self.last[src] = 0
		if sequence > self.last[src]:
			self.last[src] = sequence
			print "Chat:", nick,":", text
		#else:
			#print "Chat:", buddy,":", text, "(duplicate!)"
			
			
	def on_stdin_cb(self, sck, condition):
		line = sck.readline().strip()
		if line != "" and line != "\n":
			#print "SYS.STDIN:", line.strip()
			if line[0] == "/":
				print "this is a command"
				#self.keeper.process_command( line )
			else:
				self.send_chat_text(line)
		return True
	

if __name__ == '__main__':
	dbus.mainloop.glib.DBusGMainLoop(set_as_default = True)
	
	chat = Chat()
	
	gobject.io_add_watch(sys.stdin, gobject.IO_IN, chat.on_stdin_cb)
	
	mainloop = gobject.MainLoop()
	mainloop.run()