# !/usr/bin/env python

# Applications that use D-Bus to export a method to create a dictionary
import dbus
import dbus.service
import dbus.glib
import dbus.mainloop.glib
import gobject, traceback, sys, time
from struct import pack, unpack

import os
from cerebro.interface import *
from cerebro.common import *

def bin2mac(bin, all=False):
	if all:
		m = bin      # return all the bytes
		d = ''
	else:
		m = bin[-3:] # get the last three bytes
		d = ':'
		
	return d.join(["%02X" % ord(c) for c in m])

ATYPE = '\x01\x01'

class Debugger(dbus.service.Object, CerebroInterface):

	def __init__(self):
		CerebroInterface.__init__(self, ATYPE)
		
		self.buddyinfo = []
		self.last      = {}
		
	def _buddy_joined_cb(self, node_arr):
		for n in node_arr:
			s = "".join(chr(b) for b in n)
			print "CHAT: Node entered:", bin2mac(s)

	def _buddy_left_cb (self, node_arr):
		for n in node_arr:
			s = "".join(chr(b) for b in n)
			print "CHAT: Node left:", bin2mac(s)

	def _on_request_cb(self, src, port, req_payload):
		print "got request from:", bin2mac(src), bin2mac(port), req_payload
			
	def on_stdin_cb(self, sck, condition):
		line = sck.readline().strip()
		if line != "" and line != "\n":
			#print "SYS.STDIN:", line.strip()
			if line[0] == "/":
				self.process_command(line[1:])
			else:
				self.send_chat_text(line)
		return True
	
	def send_chat_text(self, data):
		if self.aid:
			print "chat: sending", data, len(self.aid)
			self.push_data(self.aid, data, ATYPE)
		else:
			print "Not associated with any shared activities!"
		
if __name__ == '__main__':
	dbus.mainloop.glib.DBusGMainLoop(set_as_default = True)
	
	db = Debugger()
	gobject.threads_init()
	gobject.io_add_watch(sys.stdin, gobject.IO_IN, db.on_stdin_cb)
	
	mainloop = gobject.MainLoop()
	mainloop.run()