# !/usr/bin/env python

# Applications that use D-Bus to export a method to create a dictionary
import dbus
import dbus.service
import dbus.glib
import dbus.mainloop.glib
import gobject, traceback, sys, time
from struct import pack, unpack


def bin2mac(bin, all=False):
	if all:
		m = bin      # return all the bytes
		d = ''
	else:
		m = bin[-3:] # get the last three bytes
		d = ':'
		
	return d.join(["%02X" % ord(c) for c in m])

class CerebroInterface():
	def __init__(self):
		self.aid = ''
		self.signup()
	def signup(self):
		SERVICE = 'org.olpc.cerebro'
		bus = dbus.SystemBus()
		interface = bus.get_object('org.olpc.cerebro', '/org/olpc/cerebro')
		interface.connect_to_signal("on_node_arrival", self.onNodeArrival)
		interface.connect_to_signal("received_data", self.receive_stats_data)
		interface.connect_to_signal("on_request", self.handle_stats_req)
		
		self.push_data     = interface.get_dbus_method("push_to",    SERVICE)
		self.send_file     = interface.get_dbus_method("send_file",    SERVICE)
		
class Stats(CerebroInterface):
	def __init__(self, keeper):
		
		self.keeper = keeper
		self.port = '15'
		self.rdata = {}
		self.tdata = {}
		
	def send_stats_req(self):
		dest = "" # empty destination will resolve to all neighbors
		self.keeper.request_multicast_data(dest, self.port)
			
	def handle_stats_req(self, src, port, req_payload):
		if port == self.port:
			(r, t) = self.keeper.teleporter.stats.export_to_json()
			rlen = pack("i", len(r))
			tlen = pack("i", len(t))
			res = rlen + r + tlen + t
			self.pust_to(res, src, port)
		
	def receive_stats_data(self, data, buddy, src, port, aid):
		nick = buddy['nick']
		#sequence = data[:8]
		text     = data
		#print "text", text
		rlen     = unpack("i", text[:4])[0]
		r        = text[4:4+rlen]
		tlen     = unpack("i", text[4+rlen:4+rlen+4])[0]
		t        = text[4+rlen+4:]
		#print "r", r
		#print "t", t
		if len(t) != tlen:
			print "skata:", len(t), tlen
			sys.exit()
		
		self.rdata[nick] = r
		self.tdata[nick] = t
		
	def save_stats(self):
		data = ""
		
		for b in self.rdata:
			lines = self.rdata[b].split('\n')
			
			filename = b+bin2mac(lines[0], all=True) + ".dat"
			filename = b + ".dat"
			print "will save stats file:", filename
			file = open("tests/"+filename, 'wb+')
			file.write( self.rdata[b] )
			file.close()
			
	def on_stdin_cb(self, sck, condition):
		line = sck.readline().strip()
		if line != "" and line != "\n":
			if line[0] == "/":
				command = line[1:]
				if command == "save":
					self.save_stats()
				
if __name__ == '__main__':
	dbus.mainloop.glib.DBusGMainLoop(set_as_default = True)
	
	stats = Stats()
	
	gobject.io_add_watch(sys.stdin, gobject.IO_IN, chat.on_stdin_cb)
	
	mainloop = gobject.MainLoop()
	mainloop.run()